#include <net-snmp/net-snmp-config.h>

#ifdef solaris2
#define _KMEMUSER               /* Needed by <sys/user.h> */
#include <sys/types.h>          /* helps define struct rlimit */
#endif

#if HAVE_IO_H                   /* win32 */
#include <io.h>
#endif
#if HAVE_STDLIB_H
#include <stdlib.h>
#endif
#if HAVE_UNISTD_H
#include <unistd.h>
#endif
#if HAVE_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#if HAVE_MALLOC_H
#include <malloc.h>
#endif
#include <math.h>
#include <ctype.h>
#include <sys/types.h>
#if HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#if TIME_WITH_SYS_TIME
# ifdef WIN32
#  include <sys/timeb.h>
# else
#  include <sys/time.h>
# endif
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif
#if HAVE_KVM_H
#include <kvm.h>
#endif
#if HAVE_WINSOCK_H
#include <winsock.h>
#endif

#if HAVE_DMALLOC_H
#include <dmalloc.h>
#endif

#include <net-snmp/net-snmp-includes.h>
#include <net-snmp/agent/net-snmp-agent-includes.h>

#include "struct.h"
#include "velocity_app.h"
#ifdef USING_UCD_SNMP_ERRORMIB_MODULE
#include "../ucd-snmp/errormib.h"
#else
#define setPerrorstatus(x) snmp_log_perror(x)
#endif
#include "util_funcs.h"
#include "kernel.h"
#include "errno.h"

static struct myvelocityApp *get_velocityApp_instance(struct myvelocityApp *, oid);
struct myvelocityApp  *velocityAppwatch = NULL;
int             numvelocityApps = 0;
#define ID_SIZE 300

void
init_velocity_app(void)
{

    /*
     * define the structure we're going to ask the agent to register our
     * information at 
     */
    struct variable2 extensible_velocityApp_variables[] = {
        {MIBINDEX, ASN_INTEGER, RONLY, var_extensible_velocityApp, 1, {MIBINDEX}},
        {ERRORNAME, ASN_OCTET_STR, RONLY, var_extensible_velocityApp, 1,
         {ERRORNAME}},
	{APPPID, ASN_OCTET_STR, RONLY, var_extensible_velocityApp, 1, {APPPID}}, 
        {APPMIN, ASN_INTEGER, RONLY, var_extensible_velocityApp, 1, {APPMIN}},
        {APPMAX, ASN_INTEGER, RONLY, var_extensible_velocityApp, 1, {APPMAX}},
        {APPCOUNT, ASN_INTEGER, RONLY, var_extensible_velocityApp, 1,
         {APPCOUNT}},
	{APPIDS, ASN_OCTET_STR, RONLY, var_extensible_velocityApp, 1,
         {APPIDS}},
        {ERRORFLAG, ASN_INTEGER, RONLY, var_extensible_velocityApp, 1,
         {ERRORFLAG}},
        {ERRORMSG, ASN_OCTET_STR, RONLY, var_extensible_velocityApp, 1,
         {ERRORMSG}},
    };

    /*
     * Define the OID pointer to the top of the mib tree that we're
     * registering underneath 
     */
    oid             velocityApp_variables_oid[] = { 1, 3, 6, 1, 4, 1, 15601, 6, 1 };

    /*
     * register ourselves with the agent to handle our mib tree 
     */
    REGISTER_MIB("velocity/velocity_app", extensible_velocityApp_variables, variable2,
                 velocityApp_variables_oid);

    snmpd_register_config_handler("app", velocityApp_parse_config,
                                  velocityApp_free_config,
                                  "application-name [max-num] [min-num]");
}



/*
 * Define snmpd.conf reading routines first.  They get called
 * automatically by the invocation of a macro in the proc.h file. 
 */

void
velocityApp_free_config(void)
{
    struct myvelocityApp  *ptmp, *ptmp2;

    for (ptmp = velocityAppwatch; ptmp != NULL;) {
        ptmp2 = ptmp;
        ptmp = ptmp->next;
        free(ptmp2);
    }
    velocityAppwatch = NULL;
    numvelocityApps = 0;
}

/*
 * find a give entry in the linked list associated with a proc name 
 */
static struct myvelocityApp *
get_velocityApp_by_name(char *name)
{
    struct myvelocityApp  *ptmp;

    if (name == NULL)
        return NULL;

    for (ptmp = velocityAppwatch; ptmp != NULL && strcmp(ptmp->name, name) != 0;
         ptmp = ptmp->next);
    return ptmp;
}

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

void
velocityApp_parse_config(const char *token, char *cptr)
{
    char            tmpname[STRMAX];
    char tmppid[STRMAX];
    struct myvelocityApp **procp = &velocityAppwatch;
    struct stat stat_buf;
    const char none[] = "NONE";  
    int statrc, retc;
    char *tmpstr;

    /*
     * don't allow two entries with the same name 
     */
    /* retc = strcpy(tmpstr,cptr); */
    tmpstr = cptr;
    copy_nword(tmpstr, tmpname, sizeof(tmpname));
    if (get_velocityApp_by_name(tmpname) != NULL) {
        config_perror("Already have an entry for this process.");
        return;
    }
    DEBUGMSGTL(("velocity/velocity_app", "tmpname = %s \n", tmpname));
    tmpstr = skip_not_white(tmpstr);
    tmpstr = skip_white(tmpstr);
    copy_nword(tmpstr, tmppid, sizeof(tmppid));
    retc = strcmp(tmppid, none);
    DEBUGMSGTL(("velocity/velocity_app", "tmppid = %s retc = %d\n", tmppid, retc));
    statrc = stat(tmppid, &stat_buf);
    DEBUGMSGTL(("velocity/velocity_app", "tmppid = %s statrc = %d st_mode = %d\n", tmppid, statrc, S_ISREG(stat_buf.st_mode)));
    if (retc == 0 || S_ISREG(stat_buf.st_mode)) {
	DEBUGMSGTL(("velocity/velocity_app", "Valid app string %s %s\n", tmpname, tmppid));
    }
    else {
	config_perror("Incorrect syntax for app statement. Must give either keyword NONE or absolute file.");
	return;
    }

    /*
     * skip past used ones 
     */
    while (*procp != NULL)
        procp = &((*procp)->next);

    (*procp) = (struct myvelocityApp *) calloc(1, sizeof(struct myvelocityApp));
    if (*procp == NULL)
        return;                 /* memory alloc error */
    numvelocityApps++;
    /*
     * not blank and not a comment 
     */
    copy_nword(cptr, (*procp)->name, sizeof((*procp)->name)); 
    cptr = skip_not_white(cptr);
    cptr = skip_white(cptr);
    copy_nword(cptr, (*procp)->pidfile, sizeof((*procp)->pidfile));
    DEBUGMSGTL(("velocity/velocity_app", "Read:  %s %s\n",
		(*procp)->name, (*procp)->pidfile));
    cptr = skip_not_white(cptr);
    if ((cptr = skip_white(cptr))) {
	(*procp)->max = atoi(cptr);
	cptr = skip_not_white(cptr);
	if ((cptr = skip_white(cptr)))
	    (*procp)->min = atoi(cptr);
	else
	    (*procp)->min = 0;
    } else {
	(*procp)->max = 0;
	(*procp)->min = 0;
    }
    DEBUGMSGTL(("velocity/velocity_app", "Read:  %s (%d) (%d)\n",
		(*procp)->name, (*procp)->max, (*procp)->min));
}


static struct myvelocityApp *
get_velocityApp_instance(struct myvelocityApp *proc, oid inst)
{
    int             i;

    if (proc == NULL)
        return (NULL);
    for (i = 1; (i != (int) inst) && (proc != NULL); i++)
        proc = proc->next;
    return (proc);
}
#include <dirent.h>
#include <fcntl.h>

/*
 * The routine that handles everything 
 */

u_char         *
var_extensible_velocityApp(struct variable *vp,
                    oid * name,
                    size_t * length,
                    int exact,
                    size_t * var_len, WriteMethod ** write_method)
{

    DIR *dir;
    struct myvelocityApp  *proc;
    static long     long_ret;
    static char     errmsg[ID_SIZE];

    char * cmdline = NULL;
    struct dirent *ent;
    static char fn[19];
    FILE * fd;
    char temp[ID_SIZE];
    size_t len = 0;
    ssize_t read;
    int statrc, retc;
    struct stat stat_buff;
    const char none[] = "NONE";
    char * temp2;

   if (header_simple_table
        (vp, name, length, exact, var_len, write_method, numvelocityApps))
        return (NULL);
    if ((proc = get_velocityApp_instance(velocityAppwatch, name[*length - 1]))) {
        switch (vp->magic) {
        case MIBINDEX:
            long_ret = name[*length - 1];
            return ((u_char *) (&long_ret));
	case ERRORNAME:        /* process name to check for */
	    *var_len = strlen(proc->name);
            return ((u_char *) (proc->name));
	case APPPID:
	    *var_len = strlen(proc->pidfile);
 	    DEBUGMSGTL(("velocity/velocity_app", "pid file is %s\n", proc->name ));
           return ((u_char *) (proc->pidfile));
        case APPMIN:
            long_ret = proc->min;
            return ((u_char *) (&long_ret));
        case APPMAX:
            long_ret = proc->max;
            return ((u_char *) (&long_ret));
        case APPCOUNT:
	    DEBUGMSGTL(("velocity/velocity_app", "pid file is %s\n", proc->pidfile ));
	    retc = strcmp(proc->pidfile, none);
	    DEBUGMSGTL(("velocity/velocity_app", "retc is %d, pid is %s, none is %s.\n", retc, proc->pidfile, none));
	    DEBUGMSGTL(("velocity/velocity_app", "pid length is %d, none length is %d\n",strlen(proc->pidfile), strlen(none)));
	    if ( retc == 0 || strlen(proc->pidfile) == 0 ) { 
		DEBUGMSGTL(("velocity/velocity_app", "Calling count"));
 		long_ret = sh_count_velocityApps(proc->name);
	    }
	    else {
		if (proc->pidfile != NULL) {
		    if ((fd = fopen(proc->pidfile, "r")) == NULL) break;
		    read = getline(&cmdline, &len, fd);
		    fclose(fd);
		    if (len == -1 ) break;
		    DEBUGMSGTL(("velocity/velocity_app", "Length returned is %d\n",len));
		    DEBUGMSGTL(("velocity/velocity_app", "cmdline is %s\n",cmdline));
		    cmdline = skip_white(cmdline);
		    copy_nword(cmdline, errmsg, sizeof(errmsg));
		    sprintf(fn,"/proc/%s",errmsg);
		    DEBUGMSGTL(("velocity/velocity_app", "Filename is %s\n",fn));
		    statrc = stat(fn, &stat_buff);
		    DEBUGMSGTL(("velocity/velocity_app", "statrc = %d\n",statrc));
	            if (cmdline)
	                free(cmdline);
		    if (statrc == 0 ) {
			DEBUGMSGTL(("velocity/velocity_app", "process found - %s\n", fn));
			long_ret = 1;
		    }
		    else {
			DEBUGMSGTL(("velocity/velocity_app", "process not found - %s\n",fn));
			long_ret = 0;
		    }
		}
	    }		
	    DEBUGMSGTL(("velocity/velocity_app", "Return %d\n", long_ret));
            return ((u_char *) (&long_ret));
        case APPIDS:
	    if (strcmp(proc->pidfile, none) == 0 || strlen(proc->pidfile) == 0) { 
		if ((dir = opendir("/proc")) == NULL) {
		    sprintf(fn,"-1");
		    return ((u_char *) fn);
		}
		errmsg[0] = 0;
		while (NULL != (ent = readdir(dir))) {
		    if(!(ent->d_name[0] >= '0' && ent->d_name[0] <= '9')) continue;
		    /* read /proc/XX/cmdline */
		    sprintf(fn,"/proc/%s/cmdline",ent->d_name);
		    if((fd = fopen(fn, "r")) == NULL) break;
		    read = getline(&cmdline,&len, fd);
		    fclose(fd);
		    if(read == -1) continue;
		    while(--read && !cmdline[read]);
		    while(--read) if(!cmdline[read]) cmdline[read] = ' ';
		    if(strstr(cmdline,proc->name) != NULL ) {
			snprintf(temp,sizeof(errmsg),"%s %s", errmsg, ent->d_name);
			
			strcpy(errmsg,temp);
		    }
		}
		closedir(dir);
	    }
	    else {
		DEBUGMSGTL(("velocity/velocity_app", "Appid\n"));
		if ((fd = fopen(proc->pidfile, "r")) == NULL) break;
		read = getline(&cmdline, &len, fd);
		fclose(fd);
		if (len == -1 ) break;
		cmdline = skip_white(cmdline);
		copy_nword(cmdline, errmsg, sizeof(errmsg));
		sprintf(fn,"/proc/%s",errmsg);
		DEBUGMSGTL(("velocity/velocity_app", "Filename is %s\n",fn));
		statrc = stat(fn, &stat_buff);
		DEBUGMSGTL(("velocity/velocity_app", "statrc = %d\n",statrc));
		if (statrc == 0 ) {
		    DEBUGMSGTL(("velocity/velocity_app", "process id is - %s \n", errmsg));
		}
		else {
		     DEBUGMSGTL(("velocity/velocity_app", "process id should be - %s\n", errmsg));
		     errmsg[0] = 0;
		}
	    }
	    DEBUGMSGTL(("velocity/velocity_app", "errmsg string is %s\n", errmsg));
	    if (cmdline)
	        free(cmdline);
            *var_len = strlen(errmsg);
            return ((u_char *) errmsg);
        case ERRORFLAG:
	    if (strcmp(proc->pidfile, none) == 0 || strlen(proc->pidfile) == 0) { 
		long_ret = sh_count_velocityApps(proc->name);
	    }
	    else {
	        if ((fd = fopen(proc->pidfile, "r")) == NULL) break;
		read = getline(&cmdline, &len, fd);
		fclose(fd);
		if (len == -1 ) break;
		cmdline = skip_white(cmdline);
		copy_nword(cmdline, errmsg, sizeof(errmsg));
		sprintf(fn,"/proc/%s",errmsg);
		DEBUGMSGTL(("velocity/velocity_app", "Filename is %s\n",fn));
		statrc = stat(fn, &stat_buff);
		DEBUGMSGTL(("velocity/velocity_app", "statrc = %d\n",statrc));
	        if (cmdline)
	            free(cmdline);
		if (statrc == 0 ) {
		    DEBUGMSGTL(("velocity/velocity_app", "process found - %s ", fn));
		    long_ret = 1;
		}
		else {
		    DEBUGMSGTL(("velocity/velocity_app", "process not found - %s",fn));
		    long_ret = 0;
		}
	    }		
            if (long_ret >= 0 &&
                ((proc->min && long_ret < proc->min) ||
                 (proc->max && long_ret > proc->max) ||
                 (proc->min == 0 && proc->max == 0 && long_ret < 1))) {
                long_ret = 1;
            } else {
                long_ret = 0;
            }
            return ((u_char *) (&long_ret));
        case ERRORMSG:
 	    if (strcmp(proc->pidfile, none) == 0 || strlen(proc->pidfile) == 0) { 
		long_ret = sh_count_velocityApps(proc->name);
	    }
	    else {
	        if ((fd = fopen(proc->pidfile, "r")) == NULL) break;
		read = getline(&cmdline, &len, fd);
		fclose(fd);
		if (len == -1 ) break;
		cmdline = skip_white(cmdline);
		copy_nword(cmdline, errmsg, sizeof(errmsg));
		sprintf(fn,"/proc/%s",errmsg);
		DEBUGMSGTL(("velocity/velocity_app", "Filename is %s\n",fn));
		statrc = stat(fn, &stat_buff);
		DEBUGMSGTL(("velocity/velocity_app", "statrc = %d\n",statrc));
	        if (cmdline)
	            free(cmdline);
		if (statrc == 0 ) {
		    DEBUGMSGTL(("velocity/velocity_app", "process found - %s\n ", fn));
		    long_ret = 1;
		}
		else {
		    DEBUGMSGTL(("velocity/velocity_app", "process not found - %s\n",fn));
		    long_ret = 0;
		}
	    }
            if (long_ret < 0) {
                errmsg[0] = 0;  /* catch out of mem errors return 0 count */
            } else if (proc->min && long_ret < proc->min) {
                snprintf(errmsg, sizeof(errmsg),
                        "Too few %s running (# = %d)",
                        proc->name, (int) long_ret);
            } else if (proc->max && long_ret > proc->max) {
                snprintf(errmsg, sizeof(errmsg),
                        "Too many %s running (# = %d)",
                        proc->name, (int) long_ret);
            } else if (proc->min == 0 && proc->max == 0 && long_ret < 1) {
                snprintf(errmsg, sizeof(errmsg),
                        "No %s process running.", proc->name);
            } else {
                errmsg[0] = 0;
            }
            errmsg[ sizeof(errmsg)-1 ] = 0;
            *var_len = strlen(errmsg);
            return ((u_char *) errmsg);
        }
        return NULL;
    }
    return NULL;
}


int
sh_count_velocityApps(char *velocityAppname)
{
    DIR *dir;
    char *cmdline = NULL;
    struct dirent *ent;
    FILE *fd;
    size_t len;
    char fn[18];
    int total = 0;

    if ((dir = opendir("/proc")) == NULL) return -1;
    while (NULL != (ent = readdir(dir))) {
	if(!(ent->d_name[0] >= '0' && ent->d_name[0] <= '9')) continue;
	/* read /proc/XX/cmdline */
	sprintf(fn,"/proc/%s/cmdline",ent->d_name);
	if((fd = fopen(fn, "r")) == NULL) break;
	len = getline(&cmdline,&len,fd);
	fclose(fd);
	if(len == -1) continue;
	cmdline[len] = 0;
	while(--len && !cmdline[len]);
	while(--len) if(!cmdline[len]) cmdline[len] = ' ';
	if(strstr(cmdline,velocityAppname) != NULL ) total++;
	if (cmdline)
	    free(cmdline);
    }
    closedir(dir);
    return total;
}

